# graphical display of the game world using GLFW and ModernGL
# Copyright (C) 2019  Nguyễn Gia Phong
#
# This file is part of Axuy
#
# Axuy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Axuy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Axuy.  If not, see <https://www.gnu.org/licenses/>.

__doc__ = 'Axuy graphical display using GLFW and ModernGL'
__all__ = ['DispConfig', 'Display']

from abc import abstractmethod
from collections import deque
from math import degrees, log2, radians
from random import randint
from statistics import mean
from warnings import warn

import glfw
import moderngl
import numpy as np
from PIL import Image
from pyrr import matrix44

from .misc import abspath, color, mirror
from .peer import Peer, PeerConfig
from .pico import OCTOVERTICES, SHARD_LIFE, TETRAVERTICES

CONWAY = 1.303577269034
ABRTN_MAX = 0.42069

QUAD = np.float32([-1, -1, 1, -1, -1, 1, -1, 1, 1, -1, 1, 1])
TETRAINDECIES = np.int32([0, 1, 2, 3, 1, 2, 0, 3, 2, 0, 3, 1])
OCTOINDECIES = np.int32([0, 1, 2, 0, 1, 3, 4, 0, 2, 4, 0, 3,
                         2, 1, 5, 3, 1, 5, 2, 5, 4, 3, 5, 4])

with open(abspath('shaders/map.vert')) as f: MAP_VERTEX = f.read()
with open(abspath('shaders/map.frag')) as f: MAP_FRAGMENT = f.read()
with open(abspath('shaders/pico.vert')) as f: PICO_VERTEX = f.read()
with open(abspath('shaders/pico.geom')) as f: PICO_GEOMETRY = f.read()
with open(abspath('shaders/pico.frag')) as f: PICO_FRAGMENT = f.read()

with open(abspath('shaders/tex.vert')) as f: TEX_VERTEX = f.read()
with open(abspath('shaders/sat.frag')) as f: SAT_FRAGMENT = f.read()
with open(abspath('shaders/gaussh.vert')) as f: GAUSSH_VERTEX = f.read()
with open(abspath('shaders/gaussv.vert')) as f: GAUSSV_VERTEX = f.read()
with open(abspath('shaders/gauss.frag')) as f: GAUSS_FRAGMENT = f.read()
with open(abspath('shaders/comb.frag')) as f: COMBINE_FRAGMENT = f.read()


class DispConfig(PeerConfig):
    """Graphical display configurations.

    Attributes
    ----------
    size : Tuple[int, int]
        GLFW window resolution.
    vsync : bool
        Vertical synchronization.
    zmlvl : float
        Zoom level.
    """

    def __init__(self) -> None:
        PeerConfig.__init__(self)
        self.options.add_argument(
            '--size', type=int, nargs=2, metavar=('X', 'Y'),
            help='the desired screen size (fallback: {}x{})'.format(
                *self.size))
        self.options.add_argument(
            '--vsync', action='store_true', default=None,
            help='enable vertical synchronization (fallback: {})'.format(
                self.vsync))
        self.options.add_argument(
            '--no-vsync', action='store_false', dest='vsync',
            help='disable vertical synchronization')
        self.options.add_argument(
            '--fov', type=float, metavar='DEGREES',
            help='horizontal field of view (fallback: {:})'.format(
                round(self.fov)))

    @property
    def fov(self) -> float:
        """Horizontal field of view in degrees."""
        if self.zmlvl is None: return None
        return degrees(2 ** self.zmlvl)

    @fov.setter
    def fov(self, value):
        rad = radians(value)
        if rad < 0.5:
            warn('Too narrow FOV, falling back to the minimal value.')
            self.zmlvl = -1.0
            return
        elif rad > 2:
            warn('Too wide FOV, falling back to the maximal value.')
            self.zmlvl = 1.0
            return
        self.zmlvl = log2(rad)

    def fallback(self) -> None:
        """Parse fallback configurations."""
        PeerConfig.fallback(self)
        self.size = (self.config.getint('Graphics', 'Screen width'),
                     self.config.getint('Graphics', 'Screen height'))
        self.vsync = self.config.getboolean('Graphics', 'V-sync')
        self.fov = self.config.getfloat('Graphics', 'FOV')

    def read(self, arguments):
        """Read and parse a argparse.ArgumentParser.Namespace."""
        PeerConfig.read(self, arguments)
        for option in 'size', 'vsync', 'fov':
            value = getattr(arguments, option)
            if value is not None: setattr(self, option, value)


class Display(Peer):
    """World map and camera placement.

    Parameters
    ----------
    config : DispConfig
        Display configurations.

    Attributes
    ----------
    camera : Pico
        Protagonist whose view is the camera.
    colors : Dict[Tuple[str, int], str]
        Color names of enemies.
    window : GLFW window
    zmlvl : float
        Zoom level (from ZMIN to ZMAX).
    context : moderngl.Context
        OpenGL context from which ModernGL objects are created.
    maprog : moderngl.Program
        Processed executable code in GLSL for map rendering.
    mapva : moderngl.VertexArray
        Vertex data of the map.
    prog : moderngl.Program
        Processed executable code in GLSL
        for rendering picos and their shards.
    pva : moderngl.VertexArray
        Vertex data of picos.
    sva : moderngl.VertexArray
        Vertex data of shards.
    pfilter : moderngl.VertexArray
        Vertex data for filtering highly saturated colors.
    gaussh, gaussv : moderngl.Program
        Processed executable code in GLSL for Gaussian blur.
    gausshva, gaussvva : moderngl.VertexArray
        Vertex data for Gaussian blur.
    edge : moderngl.Program
        Processed executable code in GLSL for final combination
        of the bloom effect with additional chromatic aberration
        and barrel distortion.
    combine : moderngl.VertexArray
        Vertex data for final combination of the bloom effect.
    fb, ping, pong : moderngl.Framebuffer
        Frame buffers for bloom-effect post-processing.
    fpses : Deque[float]
        FPS during the last 5 seconds to display the average.
    """

    def __init__(self, config):
        # Create GLFW window
        if not glfw.init(): raise RuntimeError('Failed to initialize GLFW')
        glfw.window_hint(glfw.CLIENT_API, glfw.OPENGL_API)
        glfw.window_hint(glfw.CONTEXT_CREATION_API, glfw.NATIVE_CONTEXT_API)
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
        glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, True)

        Peer.__init__(self, config)
        self.camera, self.colors = self.pico, {self.addr: randint(0, 5)}
        width, height = config.size
        self.zmlvl = config.zmlvl
        self.window = glfw.create_window(
            width, height, 'axuy@{}:{}'.format(*self.addr), None, None)
        if not self.window:
            glfw.terminate()
            raise RuntimeError('Failed to create GLFW window')
        self.fpses = deque()

        # Window's rendering and event-handling configuration
        glfw.set_window_icon(self.window, 1, Image.open(abspath('icon.png')))
        glfw.make_context_current(self.window)
        glfw.swap_interval(config.vsync)
        glfw.set_window_size_callback(self.window, self.resize)

        # Create OpenGL context
        self.context = context = moderngl.create_context()
        context.enable_only(context.DEPTH_TEST)

        # GLSL program and vertex array for map rendering
        self.maprog = context.program(vertex_shader=MAP_VERTEX,
                                      fragment_shader=MAP_FRAGMENT)
        mapvb = context.buffer(mirror(self.space))
        self.mapva = context.simple_vertex_array(self.maprog, mapvb, 'in_vert')

        # GLSL programs and vertex arrays for picos and shards rendering
        pvb = [(context.buffer(TETRAVERTICES), '3f', 'in_vert')]
        pib = context.buffer(TETRAINDECIES)
        svb = [(context.buffer(OCTOVERTICES), '3f', 'in_vert')]
        sib = context.buffer(OCTOINDECIES)

        self.prog = context.program(vertex_shader=PICO_VERTEX,
                                    geometry_shader=PICO_GEOMETRY,
                                    fragment_shader=PICO_FRAGMENT)
        self.pva = context.vertex_array(self.prog, pvb, pib)
        self.sva = context.vertex_array(self.prog, svb, sib)

        quad_buffer = context.buffer(QUAD)
        self.pfilter = context.simple_vertex_array(
            context.program(vertex_shader=TEX_VERTEX,
                            fragment_shader=SAT_FRAGMENT),
            quad_buffer, 'in_vert')
        self.gaussh = context.program(vertex_shader=GAUSSH_VERTEX,
                                      fragment_shader=GAUSS_FRAGMENT)
        self.gaussh['width'].value = 256
        self.gausshva = context.simple_vertex_array(
            self.gaussh, quad_buffer, 'in_vert')
        self.gaussv = context.program(vertex_shader=GAUSSV_VERTEX,
                                      fragment_shader=GAUSS_FRAGMENT)
        self.gaussv['height'].value = 256 * height / width
        self.gaussvva = context.simple_vertex_array(
            self.gaussv, quad_buffer, 'in_vert')
        self.edge = context.program(vertex_shader=TEX_VERTEX,
                                    fragment_shader=COMBINE_FRAGMENT)
        self.edge['la'].value = 0
        self.edge['tex'].value = 1
        self.combine = context.simple_vertex_array(
            self.edge, quad_buffer, 'in_vert')

        size, table = (width, height), (256, height * 256 // width)
        self.fb = context.framebuffer(context.texture(size, 4),
                                      context.depth_renderbuffer(size))
        self.fb.color_attachments[0].use(1)
        self.ping = context.framebuffer(context.texture(table, 3))
        self.pong = context.framebuffer(context.texture(table, 3))

    def resize(self, window, width, height):
        """Update viewport on resize."""
        context = self.context
        context.viewport = 0, 0, width, height
        self.gaussv['height'].value = 256 * height / width

        self.fb.depth_attachment.release()
        for fb in (self.fb, self.ping, self.pong):
            for texture in fb.color_attachments: texture.release()
            fb.release()

        size, table = (width, height), (256, height * 256 // width)
        self.fb = context.framebuffer(context.texture(size, 4),
                                      context.depth_renderbuffer(size))
        self.fb.color_attachments[0].use(1)
        self.ping = context.framebuffer(context.texture(table, 3))
        self.pong = context.framebuffer(context.texture(table, 3))

    @property
    def width(self) -> int:
        """Viewport width."""
        return self.context.viewport[2]

    @property
    def height(self) -> int:
        """Viewport height."""
        return self.context.viewport[3]

    @property
    def health(self) -> float:
        """Camera relative health point."""
        return self.camera.health

    @property
    def pos(self) -> np.float32:
        """Camera position in a NumPy array."""
        return self.camera.pos

    @property
    def postr(self) -> str:
        """Pretty camera position representation."""
        return '[{:4.1f} {:4.1f} {:3.1f}]'.format(*self.pos)

    @property
    def right(self) -> np.float32:
        """Camera right direction."""
        return self.camera.rot[0]

    @property
    def upward(self) -> np.float32:
        """Camera upward direction."""
        return self.camera.rot[1]

    @property
    def forward(self) -> np.float32:
        """Camera forward direction."""
        return self.camera.rot[2]

    @property
    def is_running(self) -> bool:
        """GLFW window status."""
        return not glfw.window_should_close(self.window)

    @property
    def fov(self) -> float:
        """Horizontal field of view in degrees."""
        return degrees(2 ** self.zmlvl)

    @property
    def visibility(self) -> np.float32:
        """Camera visibility."""
        return np.float32(3240 / (self.fov + 240))

    @property
    def fpstr(self) -> str:
        """Pretty string for displaying average FPS."""
        # Average over 5 seconds, like how glxgears do it, but less efficient
        while len(self.fpses) > mean(self.fpses) * 5 > 0: self.fpses.pop()
        return '{} fps'.format(round(mean(self.fpses)))

    def get_time(self) -> float:
        """Return the current time in seconds."""
        return glfw.get_time()

    def prender(self, obj, va, col, bright):
        """Render the obj and its images in bounded 3D space."""
        self.prog['rot'].write(
            matrix44.create_from_matrix33(obj.rot, dtype=np.float32))
        self.prog['pos'].write(obj.pos)
        self.prog['color'].write(color(col, bright))
        va.render(moderngl.TRIANGLES)

    def render_pico(self, pico):
        """Render pico and its images in bounded 3D space."""
        self.prender(pico, self.pva, self.colors[pico.addr], pico.health)

    def render_shard(self, shard):
        """Render shard and its images in bounded 3D space."""
        self.prender(shard, self.sva,
                     self.colors[shard.addr], shard.power/SHARD_LIFE)

    def add_pico(self, address):
        """Add pico from given address."""
        Peer.add_pico(self, address)
        self.colors[address] = randint(0, 5)

    def render(self) -> None:
        """Render the scene before post-processing."""
        visibility = self.visibility
        projection = matrix44.create_perspective_projection(
            self.fov, self.width/self.height, 3E-3, visibility,
            dtype=np.float32)
        view = matrix44.create_look_at(
            self.pos, self.pos+self.forward, self.upward, dtype=np.float32)
        vp = view @ projection

        # Render map
        self.maprog['visibility'].value = visibility
        self.maprog['mvp'].write(vp)
        self.mapva.render(moderngl.TRIANGLES)

        # Render picos and shards
        self.prog['visibility'].value = visibility
        self.prog['camera'].write(self.pos)
        self.prog['vp'].write(vp)
        for pico in self.picos.values():
            for shard in pico.shards.values(): self.render_shard(shard)
            if pico is not self.camera: self.render_pico(pico)

    def update(self) -> None:
        """Update and render the map."""
        # Update states
        Peer.update(self)
        self.fpses.appendleft(self.fps)

        # Render to framebuffer
        self.fb.use()
        self.fb.clear()
        self.render()
        self.fb.color_attachments[0].use()
        self.ping.use()
        self.ping.clear()
        self.pfilter.render(moderngl.TRIANGLES)
        self.ping.color_attachments[0].use()

        # Gaussian blur
        self.pong.use()
        self.pong.clear()
        self.gausshva.render(moderngl.TRIANGLES)
        self.pong.color_attachments[0].use()
        self.ping.use()
        self.ping.clear()
        self.gaussvva.render(moderngl.TRIANGLES)
        self.ping.color_attachments[0].use()

        # Combine for glow effect, chromatic aberration and barrel distortion
        self.context.screen.use()
        self.context.clear()
        if self.camera.dead:
            abrtn = ABRTN_MAX
        else:
            abrtn = min(ABRTN_MAX, (self.fov*self.health) ** -CONWAY)
        self.edge['abrtn'].value = abrtn
        self.edge['zoom'].value = (self.zmlvl + 1.0) / 100
        self.combine.render(moderngl.TRIANGLES)
        glfw.swap_buffers(self.window)
        glfw.set_window_title(self.window, '{} - axuy@{}:{} ({})'.format(
            self.postr, *self.addr, self.fpstr))

    @abstractmethod
    def control(self) -> None:
        """Poll resizing and closing events."""
        glfw.poll_events()

    def __exit__(self, exc_type, exc_value, traceback):
        Peer.__exit__(self, exc_type, exc_value, traceback)
        glfw.terminate()



# handling of user control using GLFW
# Copyright (C) 2019  Nguyễn Gia Phong
#
# This file is part of Axuy
#
# Axuy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Axuy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Axuy.  If not, see <https://www.gnu.org/licenses/>.

__doc__ = 'Axuy handling of user control using GLFW'
__all__ = ['CtlConfig', 'Control']

from cmath import polar
from re import IGNORECASE, match
from warnings import warn

import glfw
import numpy

from .display import DispConfig, Display

CONTROL_ALIASES = (('Move left', 'left'), ('Move right', 'right'),
                   ('Move forward', 'forward'), ('Move backward', 'backward'),
                   ('Primary', '1st'), ('Secondary', '2nd'))
MOUSE_PATTERN = 'MOUSE_BUTTON_[1-{}]'.format(glfw.MOUSE_BUTTON_LAST + 1)
INVALID_CONTROL_ERR = '{}: {} is not recognized as a valid control key'
GLFW_VER_WARN = 'Your GLFW version appear to be lower than 3.3, '\
                'which might cause stuttering camera rotation.'
ZMIN, ZMAX = -1.0, 1.0


class CtlConfig(DispConfig):
    """User control configurations.

    Attributes
    ----------
    key, mouse : Dict[str, int]
        Input control.
    mouspeed : float
        Relative camera rotational speed.
    zmspeed : float
        Zoom speed, in scroll steps per zoom range.
    """

    def __init__(self) -> None:
        DispConfig.__init__(self)
        self.options.add_argument(
            '--mouse-speed', type=float, dest='mouspeed',
            help='camera rotational speed (fallback: {:.1f})'.format(
                self.__mouspeed))
        self.options.add_argument(
            '--zoom-speed', type=float, dest='zmspeed',
            help='zoom speed (fallback: {:.1f})'.format(self.zmspeed))

    @property
    def mouspeed(self) -> float:
        """Relative mouse speed."""
        # Standard to radians per inch for a 800 DPI mouse, at FOV of 60
        return self.__mouspeed / 800

    @mouspeed.setter
    def mouspeed(self, value: float) -> None:
        self.__mouspeed = value

    def fallback(self) -> None:
        """Parse fallback configurations."""
        DispConfig.fallback(self)
        self.mouspeed = self.config.getfloat('Control', 'Mouse speed')
        self.zmspeed = self.config.getfloat('Control', 'Zoom speed')
        self.key, self.mouse = {}, {}
        for cmd, alias in CONTROL_ALIASES:
            i = self.config.get('Control', cmd)
            if match(MOUSE_PATTERN, i, flags=IGNORECASE):
                self.mouse[alias] = getattr(glfw, i.upper())
                continue
            try:
                self.key[alias] = getattr(glfw, 'KEY_{}'.format(i.upper()))
            except AttributeError:
                raise ValueError(INVALID_CONTROL_ERR.format(cmd, i))

    def read(self, arguments):
        """Read and parse a argparse.ArgumentParser.Namespace."""
        DispConfig.read(self, arguments)
        for option in 'fov', 'mouspeed', 'zmspeed':
            value = getattr(arguments, option)
            if value is not None: setattr(self, option, value)


class Control(Display):
    """User control.

    Parameters
    ----------
    config : CtlConfig
        User control configurations.

    Attributes
    ----------
    key, mouse : Dict[str, int]
        Input control.
    zmspeed : float
        Scroll steps per zoom range.
    mouspeed : float
        Relative camera rotational speed.
    """

    def __init__(self, config):
        Display.__init__(self, config)
        self.key, self.mouse = config.key, config.mouse
        self.mouspeed = config.mouspeed
        self.zmspeed = config.zmspeed

        glfw.set_input_mode(self.window, glfw.CURSOR, glfw.CURSOR_DISABLED)
        glfw.set_input_mode(self.window, glfw.STICKY_KEYS, True)
        glfw.set_cursor_pos_callback(self.window, self.look)
        glfw.set_scroll_callback(self.window, self.zoom)
        glfw.set_mouse_button_callback(self.window, self.shoot)
        try:
            if glfw.raw_mouse_motion_supported():
                glfw.set_input_mode(self.window, glfw.RAW_MOUSE_MOTION, True)
        except AttributeError:
            warn(GLFW_VER_WARN, category=RuntimeWarning)

    def look(self, window, xpos, ypos):
        """Look according to cursor position.

        Present as a callback for GLFW CursorPos event.
        """
        center = numpy.array(glfw.get_window_size(window)) / 2
        glfw.set_cursor_pos(window, *center)
        yaw, pitch = (center - [xpos, ypos]) * self.mouspeed * 2**self.zmlvl
        self.camera.rotate(*polar(complex(yaw, pitch)))

    def zoom(self, window, xoffset, yoffset):
        """Adjust FOV according to vertical scroll."""
        self.zmlvl += yoffset * 2 / self.zmspeed
        self.zmlvl = max(self.zmlvl, ZMIN)
        self.zmlvl = min(self.zmlvl, ZMAX)

    def shoot(self, window, button, action, mods):
        """Shoot on click.

        Present as a callback for GLFW MouseButton event.
        """
        if action == glfw.PRESS:
            if button == self.mouse['1st']:
                self.camera.shoot()
            elif button == self.mouse['2nd']:
                self.camera.shoot(backward=True)

    def is_pressed(self, *keys) -> bool:
        """Return whether given keys are pressed."""
        return any(glfw.get_key(self.window, k) == glfw.PRESS for k in keys)

    def control(self) -> None:
        """Handle events controlling the protagonist."""
        Display.control(self)
        right, upward, forward = 0, 0, 0
        if self.is_pressed(self.key['forward']): forward += 1
        if self.is_pressed(self.key['backward']): forward -= 1
        if self.is_pressed(self.key['left']): right -= 1
        if self.is_pressed(self.key['right']): right += 1
        self.pico.update(right, upward, forward)



# p2p networking
# Copyright (C) 2019  Nguyễn Gia Phong
#
# This file is part of Axuy
#
# Axuy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Axuy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Axuy.  If not, see <https://www.gnu.org/licenses/>.

__doc__ = 'Axuy peer'
__all__ = ['PeerConfig', 'Peer']
__version__ = '0.0.10'

from abc import ABC, abstractmethod
from argparse import ArgumentParser, FileType, RawTextHelpFormatter
from configparser import ConfigParser
from os.path import join as pathjoin, pathsep
from pickle import dumps, loads
from queue import Empty, Queue
from socket import SO_REUSEADDR, SOCK_DGRAM, SOL_SOCKET, socket
from sys import stdout
from threading import Thread
from typing import Iterator, Tuple

from appdirs import AppDirs

from .misc import abspath, mapgen, mapidgen
from .pico import Pico

SETTINGS = abspath('settings.ini')


class PeerConfig:
    """Networking configurations.

    Attributes
    ----------
    config : ConfigParser
        INI configuration file parser.
    options : ArgumentParser
        Command-line argument parser.
    host : str
        Host to bind the peer to.
    port : int
        Port to bind the peer to.
    seeder : str
        Address of the peer that created the map.
    """

    def __init__(self) -> None:
        dirs = AppDirs(appname='axuy', appauthor=False, multipath=True)
        parents = dirs.site_config_dir.split(pathsep)
        parents.append(dirs.user_config_dir)
        filenames = [pathjoin(parent, 'settings.ini') for parent in parents]

        # Parse configuration files
        self.config = ConfigParser()
        self.config.read(SETTINGS)
        self.config.read(filenames)
        self.fallback()

        # Parse command-line arguments
        self.options = ArgumentParser(usage='%(prog)s [options]',
                                      formatter_class=RawTextHelpFormatter)
        self.options.add_argument('-v', '--version', action='version',
                                  version='Axuy {}'.format(__version__))
        self.options.add_argument(
            '--write-config', nargs='?', const=stdout, type=FileType('w'),
            metavar='PATH', dest='cfgout',
            help='write default config to PATH (fallback: stdout) and exit')
        self.options.add_argument(
            '-c', '--config', metavar='PATH',
            help='location of the configuration file (fallback: {})'.format(
                pathsep.join(filenames)))
        self.options.add_argument(
            '--host',
            help='host to bind this peer to (fallback: {})'.format(self.host))
        self.options.add_argument(
            '-p', '--port', type=int,
            help='port to bind this peer to (fallback: {})'.format(self.port))
        self.options.add_argument(
            '-s', '--seeder', metavar='ADDRESS',
            help='address of the peer that created the map')

    def fallback(self) -> None:
        """Parse fallback configurations."""
        self.host = self.config.get('Peer', 'Host')
        self.port = self.config.getint('Peer', 'Port')

    # Fallback to None when attribute is missing
    def __getattr__(self, name): return None

    @property
    def seeder(self) -> Tuple[str, int]:
        """Seeder address."""
        return self.__seed

    @seeder.setter
    def seeder(self, value: str) -> None:
        host, port = value.split(':')
        self.__seed = host, int(port)

    def read(self, arguments):
        """Read and parse a argparse.ArgumentParser.Namespace."""
        for option in 'host', 'port', 'seeder':
            value = getattr(arguments, option)
            if value is not None: setattr(self, option, value)

    def parse(self) -> None:
        """Parse all configurations."""
        args = self.options.parse_args()
        if args.cfgout is not None:
            with open(SETTINGS) as f: args.cfgout.write(f.read())
            args.cfgout.close()
            exit()
        if args.config:     # is neither None nor empty
            self.config.read(args.config)
            self.fallback()
        self.read(args)


class Peer(ABC):
    """Axuy peer.

    Parameters
    ----------
    config : PeerConfig
        Networking configurations.

    Attributes
    ----------
    sock : socket
        UDP socket for exchanging instantaneous states with other peers.
    addr : Tuple[str, int]
        Own's address.
    q : Queue[Tuple[bytes, Tuple[str, int]]]
        Queue of (data, addr), where addr is the address of the peer
        who sent the raw data.
    peers : List[Tuple[str, int]]
        Addresses of connected peers.
    mapid : List[int]
        Permutation of map building blocks.
    space : numpy.ndarray of shape (12, 12, 9) of bools
        3D array of occupied space.
    pico : Pico
        Protagonist.
    picos : Dict[Tuple[str, int], Pico]
        All picos present in the map.
    last_time : float
        Timestamp of the previous update.
    """

    def __init__(self, config):
        self.sock = socket(type=SOCK_DGRAM)     # UDP
        self.sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.sock.bind((config.host, config.port))
        self.addr = self.sock.getsockname()
        self.q = Queue()

        if config.seeder is None:
            self.mapid, self.peers = mapidgen(), []
        else:
            client = socket()
            client.connect(config.seeder)
            self.mapid, self.peers = loads(client.recv(1024))

        self.space = mapgen(self.mapid)
        self.pico = Pico(self.addr, self.space)
        self.picos = {self.addr: self.pico}
        self.last_time = self.get_time()

    def __enter__(self): return self

    @property
    @abstractmethod
    def is_running(self) -> bool:
        """Peer status."""

    @property
    def ready(self) -> Iterator[Tuple[bytes, Tuple[str, int]]]:
        """Iterator of (data, addr) that can be used without waiting,
        where addr is the address of the peer who sent the data.
        """
        while True:
            try:
                yield self.q.get_nowait()
            except Empty:
                break
            else:
                self.q.task_done()

    @property
    def fps(self) -> float:
        """Current loop rate."""
        return self.pico.fps

    @fps.setter
    def fps(self, fps: float) -> None:
        self.pico.fps = fps

    def serve(self) -> None:
        """Initiate other peers."""
        with socket() as server:    # TCP server
            server.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
            server.bind(self.addr)
            server.listen(7)
            print('Axuy is listening at {}:{}'.format(*self.addr))
            while self.is_running:
                conn, addr = server.accept()
                conn.send(dumps((self.mapid, self.peers+[self.addr])))
                conn.close()
            server.close()

    def pull(self) -> None:
        """Receive other peers' states."""
        while self.is_running: self.q.put(self.sock.recvfrom(1 << 16))
        while not self.q.empty():
            self.q.get()
            self.q.task_done()

    @abstractmethod
    def get_time(self) -> float:
        """Return the current time in seconds."""

    def add_pico(self, address):
        """Add pico from given address."""
        self.picos[address] = Pico(address, self.space)

    def sync(self) -> None:
        """Synchronize states received from other peers."""
        for data, addr in self.ready:
            if addr not in self.picos:
                self.peers.append(addr)
                self.add_pico(addr)
            self.picos[addr].sync(*loads(data))

    def push(self) -> None:
        """Push states to other peers."""
        shards = {i: (s.pos, s.rot, s.power)
                  for i, s in self.pico.shards.items()}
        data = dumps([self.pico.health, self.pico.pos, self.pico.rot, shards])
        for peer in self.peers: self.sock.sendto(data, peer)

    @abstractmethod
    def control(self) -> None:
        """Control the protagonist."""
        self.pico.update()  # just a reminder that this needs to be called

    def update(self) -> None:
        """Update internal states and send them to other peers."""
        next_time = self.get_time()
        self.fps = 1 / (next_time-self.last_time)
        self.last_time = next_time

        self.sync()
        self.control()
        picos = list(self.picos.values())
        for pico in picos:
            shards = {}
            for index, shard in pico.shards.items():
                shard.update(self.fps, picos)
                if shard.power: shards[index] = shard
            pico.shards = shards
        self.push()

    def run(self) -> None:
        """Start main loop."""
        Thread(target=self.serve, daemon=True).start()
        Thread(target=self.pull, daemon=True).start()
        while self.is_running: self.update()

    def __exit__(self, exc_type, exc_value, traceback):
        self.q.join()
        self.sock.close()



# main loop
# Copyright (C) 2019  Nguyễn Gia Phong
#
# This file is part of Axuy
#
# Axuy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Axuy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Axuy.  If not, see <https://www.gnu.org/licenses/>.

__doc__ = 'Axuy main loop'
__all__ = ['main']

from .control import Control, CtlConfig


def main():
    """Parse arguments and start main loop."""
    config = CtlConfig()
    config.parse()
    with Control(config) as peer: peer.run()


if __name__ == '__main__': main()



# game characters and bullets
# Copyright (C) 2019  Nguyễn Gia Phong
#
# This file is part of Axuy
#
# Axuy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Axuy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Axuy.  If not, see <https://www.gnu.org/licenses/>.

__doc__ = 'Axuy module for character and bullet class'
__all__ = ['TETRAVERTICES', 'OCTOVERTICES', 'RPICO', 'RSHARD', 'RCOLL', 'INV',
           'PICO_SPEED', 'SHARD_SPEED', 'SHARD_LIFE', 'RPS', 'Pico', 'Shard']

from itertools import combinations
from math import acos, atan2, log10, pi, sqrt
from random import random

import numpy as np
from numpy.linalg import norm

from .misc import normalized, placeable, rot33

TETRAVERTICES = np.float32([[0, sqrt(8), -1], [sqrt(6), -sqrt(2), -1],
                            [0, 0, 3], [-sqrt(6), -sqrt(2), -1]]) / 18
OCTOVERTICES = np.float32([(a + b) / 2.0
                           for a, b in combinations(TETRAVERTICES, 2)])
RPICO = norm(TETRAVERTICES[0])
RSHARD = norm(OCTOVERTICES[0])
RCOLL = RPICO * 2/3

INVX = np.float32([[-1, 0, 0], [0, 1, 0], [0, 0, 1]])
INVY = np.float32([[1, 0, 0], [0, -1, 0], [0, 0, 1]])
INVZ = np.float32([[1, 0, 0], [0, 1, 0], [0, 0, -1]])
INV = {'x': INVX, 'y': INVY, 'z': INVZ}

PICO_SPEED = 1 + sqrt(5)        # unit/s
SHARD_SPEED = PICO_SPEED * 2    # unit/s
SHARD_LIFE = 3  # bounces
RPS = pi    # rounds per second


class Shard:
    """Fragment broken or shot out of a Pico, which is a regular
    octahedron whose circumscribed sphere's radius is RSHARD.

    Parameters
    ----------
    address : Tuple[str, int]
        IP address (host, port).
    space : np.ndarray of shape (12, 12, 9) of bools
        3D array of occupied space.
    position : iterable of length 3 of floats
        Position.
    rotation : np.ndarray of shape (3, 3) of np.float32
        Rotational matrix.
    power : int, optional
        Relative destructive power.

    Attributes
    ----------
    addr : Tuple[str, int]
        IP address (host, port).
    power : int
        Relative destructive power.
    space : np.ndarray of shape (12, 12, 9) of bools
        3D array of occupied space.
    x, y, z : floats
        Position.
    rot : np.ndarray of shape (3, 3) of np.float32
        Rotational matrix.
    """
    def __init__(self, address, space, position, rotation, power=SHARD_LIFE):
        self.addr = address
        self.power = power
        self.space = space
        self.pos = position
        self.rot = rotation

    @property
    def pos(self) -> np.float32:
        """Position in a NumPy array."""
        return np.float32([self.x, self.y, self.z])

    @pos.setter
    def pos(self, position):
        x, y, z = position
        self.x = x % 12
        self.y = y % 12
        self.z = z % 9

    @property
    def forward(self) -> np.float32:
        """Direction in a NumPy array."""
        return self.rot[-1]

    def should_bounce(self, x=None, y=None, z=None) -> bool:
        """Return whether it should bounce at (x, y, z)."""
        if x is None: x = self.x
        if y is None: y = self.y
        if z is None: z = self.z
        return not placeable(self.space, x, y, z, r=RSHARD)

    def update(self, fps, picos):
        """Update states."""
        bounced = False
        for axis, value in zip('xyz', self.pos+self.forward/fps*SHARD_SPEED):
            if self.should_bounce(**{axis: value}):
                self.rot = self.rot @ INV[axis]
                bounced = True
        self.pos += self.forward / fps * SHARD_SPEED
        self.power -= bounced

        for pico in picos:
            if norm(pico.pos - self.pos) < RCOLL:
                pico.health -= self.power / SHARD_LIFE / RPS
                self.power = 0

    def sync(self, position, rotation, power) -> None:
        """Synchronize states received from other peers."""
        self.pos, self.rot, self.power = position, rotation, power


class Pico:
    """Game character, which is represented as a regular tetrahedron
    whose circumscribed sphere's radius is RPICO.

    Parameters
    ----------
    address : Tuple[str, int]
        IP address (host, port).
    space : np.ndarray of shape (12, 12, 9) of bools
        3D array of occupied space.
    health : float, optional
        Pico relative health.
    position : iterable of length 3 of floats, optional
        Position.
    rotation : np.ndarray of shape (3, 3) of np.float32, optional
        Rotational matrix.

    Attributes
    ----------
    addr : Tuple[str, int]
        IP address (host, port).
    space : np.ndarray of shape (12, 12, 9) of bools
        3D array of occupied space.
    health : float
        Pico relative health.
    x, y, z : floats
        Position.
    rot : np.ndarray of shape (3, 3) of np.float32
        Rotational matrix.
    shards : Dict[int, Shard]
        Active shards.
    recoil_u : np.ndarray of length 3 of np.float32
        Recoil direction (unit vector).
    recoil_t : float
        Recoil time left in seconds.
    fps : float
        Currently rendered frames per second.
    """
    def __init__(self, address, space,
                 health=1.0, position=None, rotation=None):
        self.addr = address
        self.space = space
        self.health = health

        if position is None:
            x, y, z = random()*12, random()*12, random()*9
            while not self.placeable(x, y, z):
                x, y, z = random()*12, random()*12, random()*9
            self.x, self.y, self.z = x, y, z
        else:
            self.x, self.y, self.z = position

        if rotation is None:
            self.rot = INVZ
            self.rotate(random()*pi*2, random()*pi*2)
        else:
            self.rot = rotation

        self.shards = {}
        self.recoil_u, self.recoil_t = np.float32([0, 0, 0]), 0.0
        self.fps = 60.0

    @property
    def dead(self) -> bool:
        """Whether the pico is dead."""
        return self.health < 0

    @property
    def forward(self) -> np.float32:
        """Direction in a NumPy array."""
        return self.rot[-1]

    @property
    def pos(self) -> np.float32:
        """Position in a NumPy array."""
        return np.float32([self.x, self.y, self.z])

    @pos.setter
    def pos(self, position):
        self.x, self.y, self.z = position

    def sync(self, health, position, rotation, shards):
        """Synchronize states received from other peers."""
        self.health, self.pos, self.rot = health, position, rotation
        for i, t in shards.items():
            pos, rot, power = t
            try:
                self.shards[i].sync(pos, rot, power)
            except KeyError:
                self.shards[i] = Shard(self.addr, self.space, pos, rot, power)

    def placeable(self, x=None, y=None, z=None) -> bool:
        """Return whether it can be placed at (x, y, z)."""
        if x is None: x = self.x
        if y is None: y = self.y
        if z is None: z = self.z
        return placeable(self.space, x, y, z, RPICO)

    def rotate(self, magnitude, direction):
        """Rotate by the given magnitude and direction."""
        self.rot = rot33(magnitude, direction) @ self.rot

    def lookat(self, target):
        """Look at the given target."""
        # The matrix multiplication is in the following order
        # because matrices are perceived differently by numpy and glsl.
        right, upward, forward = normalized(*(self.rot @ (target - self.pos)))
        # I don't understand why we need to flip
        # the right coordinate here, but since it works...
        self.rotate(acos(forward), atan2(upward, -right))

    def update(self, right=0, upward=0, forward=0):
        """Recover health point and try to move in the given direction."""
        if self.dead: return self.__init__(self.addr, self.space)   # respawn
        dt = 1.0 / self.fps
        self.health = min(1.0, self.health + log10(self.health+1)*dt)

        direction = normalized(right, upward, forward) @ self.rot
        if self.recoil_t:
            direction += self.recoil_u * self.recoil_t * RPS
            self.recoil_t = max(self.recoil_t - dt, 0.0)
        x, y, z = self.pos + direction*dt*PICO_SPEED
        if self.placeable(x=x): self.x = x % 12
        if self.placeable(y=y): self.y = y % 12
        if self.placeable(z=z): self.z = z % 9

    def add_shard(self, pos, rot):
        """Add a shard at pos with rotation rot."""
        self.shards[max(self.shards, default=0) + 1] = Shard(
            self.addr, self.space, pos-self.recoil_u*RPICO, rot)

    def shoot(self, backward=False):
        """Shoot in the forward direction unless specified otherwise."""
        if self.recoil_t or self.dead: return
        self.recoil_t = 1.0 / RPS
        if backward:
            self.recoil_u = self.forward
            self.add_shard(-self.pos, -self.rot)
        else:
            self.recoil_u = -self.forward
            self.add_shard(self.pos, self.rot)



# package initialization
# Copyright (C) 2019  Nguyễn Gia Phong
#
# This file is part of Axuy
#
# Axuy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Axuy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Axuy.  If not, see <https://www.gnu.org/licenses/>.

"""Axuy is a minimalist peer-to-peer first-person shooter.

This package provides abstractions for writing custom front-ends and AIs.
All classes and helper functions are exposed at the package level.

Some superclasses may define abstract methods which must be overridden
in derived classes.  Subclasses only document newly introduced attributes.
"""

from .control import *
from .display import *
from .misc import *
from .peer import *
from .pico import *

__all__ = (misc.__all__ + pico.__all__ + peer.__all__
           + display.__all__ + control.__all__)
__version__ = peer.__version__



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['axuy']

package_data = \
{'': ['*'], 'axuy': ['shaders/*']}

install_requires = \
['appdirs', 'glfw>=1.8', 'moderngl', 'numpy', 'Pillow', 'pyrr', 'setuptools']

entry_points = \
{'console_scripts': ['axuy = axuy.__main__:main']}

setup(name='axuy',
      version='0.0.10',
      description='Axuy is a minimalist peer-to-peer first-person shooter.',
      author='Nguyễn Gia Phong',
      author_email='mcsinyx@disroot.org',
      url='https://github.com/McSinyx/axuy',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



# miscellaneous functions
# Copyright (C) 2019  Nguyễn Gia Phong
#
# This file is part of Axuy
#
# Axuy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Axuy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Axuy.  If not, see <https://www.gnu.org/licenses/>.

__doc__ = 'Axuy miscellaneous functions'
__all__ = ['abspath', 'color', 'twelve', 'nine', 'indexify', 'mapidgen',
           'mapgen', 'neighbors', 'mirror', 'normalized', 'rot33', 'placeable']

from itertools import (chain, combinations_with_replacement,
                       permutations, product)
from random import choices, shuffle
from typing import Iterator, List, Tuple

import numpy
from numpy.linalg import norm
from pkg_resources import resource_filename
from pyrr import matrix33

OXY = numpy.float32([[0, 0, 0], [1, 0, 0], [1, 1, 0],
                     [1, 1, 0], [0, 1, 0], [0, 0, 0]])
OYZ = numpy.float32([[0, 0, 0], [0, 1, 0], [0, 1, 1],
                     [0, 1, 1], [0, 0, 1], [0, 0, 0]])
OZX = numpy.float32([[0, 0, 0], [1, 0, 0], [1, 0, 1],
                     [1, 0, 1], [0, 0, 1], [0, 0, 0]])
AXIS = numpy.float32([0, -1, 0])

NEIGHBORS = set(chain.from_iterable(
    map(permutations, combinations_with_replacement((-1, 0, 1), 3))))
# map.npy is generated by ../tools/mapgen
SPACE = numpy.load(resource_filename('axuy', 'map.npy'))
COLORS = tuple(map(numpy.float32, permutations((1.0, 0.5, 0.0))))


def abspath(resource_name) -> str:
    """Return a true filesystem path for the specified resource."""
    return resource_filename('axuy', resource_name)


def color(code, value) -> numpy.float32:
    """Return NumPy float32 array of RGB colors from color name."""
    return COLORS[code] * (value + 1) * 0.5


def twelve(x) -> int:
    """Shorthand for int(x % 12)."""
    return int(x % 12)


def nine(x) -> int:
    """Shorthand for int(x % 9)."""
    return int(x % 9)


def indexify(iterable) -> Tuple[int, int, int]:
    """Return a tuple of int to be used as indices of the map's space."""
    x, y, z = iterable
    return twelve(x), twelve(y), nine(y)


def mapidgen(replacement=False) -> List[int]:
    """Return a randomly generated map ID."""
    mapid = list(range(48))
    if replacement: return choices(mapid, k=48)
    shuffle(mapid)
    return mapid


def mapgen(mapid):
    """Return the NumPy array of shape (12, 12, 9) of bools
    generated from the given ID.
    """
    base = numpy.stack([SPACE[i] for i in mapid]).reshape(4, 4, 3, 3, 3, 3)
    space = numpy.zeros([12, 12, 9], dtype=bool)
    for (i, j, k, x, y, z), occupied in numpy.ndenumerate(base):
        if occupied: space[i*3 + x][j*3 + y][k*3 + z] = 1
    return space


def neighbors(x, y, z) -> Iterator[Tuple[float, float, float]]:
    """Return a generator of coordinates of images point (x, y, z)
    in neighbor universes.
    """
    for i, j, k in NEIGHBORS: yield x + i*12, y + j*12, z + k*9


def mirror(space) -> numpy.float32:
    """Return the vertices to render the mirrored space."""
    vertices = []
    for (x, y, z), occupied in numpy.ndenumerate(space):
        if space[x][y][z-1] ^ occupied:
            vertices.extend(i+j for i, j in product(neighbors(x, y, z), OXY))
        if space[x-1][y][z] ^ occupied:
            vertices.extend(i+j for i, j in product(neighbors(x, y, z), OYZ))
        if space[x][y-1][z] ^ occupied:
            vertices.extend(i+j for i, j in product(neighbors(x, y, z), OZX))
    return numpy.stack(vertices).astype(numpy.float32)


def normalized(*vector) -> numpy.float32:
    """Return normalized vector as a NumPy array of float32."""
    v = numpy.float32(vector)
    if not v.any(): return v
    return v / norm(v)


def placeable(space, x, y, z, r=0) -> bool:
    """Return whether a sphere of radius r
    can be placed at (x, y, z) in given space.
    """
    return not any(space[i][j][k] for i, j, k in product(
        {twelve(x-r), twelve(x), twelve(x+r)},
        {twelve(y-r), twelve(y), twelve(y+r)},
        {nine(z-r), nine(z), nine(z+r)}))


def rot33(magnitude, direction) -> numpy.float32:
    """Return the 3x3 matrix of float32 which rotates
    by the given magnitude and direction.
    """
    return matrix33.create_from_axis_rotation(
        AXIS.dot(matrix33.create_from_z_rotation(direction)),
        magnitude,
        dtype=numpy.float32)
